/* id_pt.c: Interdata paper tape reader

   Copyright (c) 2000-2001, Robert M. Supnik

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   ROBERT M SUPNIK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of Robert M Supnik shall not
   be used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from Robert M Supnik.

   pt			paper tape reader and punch
*/

#include "id_defs.h"
#include <ctype.h>

/* Device definitions */

#define PTR		0				/* unit subscripts */
#define PTP		1

#define STA_OVR		0x80				/* overrun */
#define STA_NMTN	0x10				/* no motion */
#define STA_MASK	(STA_BSY | STA_OVR | STA_DU)	/* static bits */
#define SET_EX		(STA_OVR | STA_NMTN)		/* set EX */

#define CMD_V_RUN	4				/* run/stop */
#define CMD_V_SLEW	2				/* slew/step */
#define CMD_V_RD	0				/* read/write */

extern uint32 int_req[INTSZ], int_enb[INTSZ];

uint32 pt_run = 0, pt_slew = 0;				/* ptr modes */
uint32 pt_rd = 1, pt_chp = 0;				/* pt state */
uint32 pt_arm = 0;					/* int arm */
uint32 pt_sta = STA_BSY;				/* status */
uint32 ptr_stopioe = 0, ptp_stopioe = 0;		/* stop on error */

DEVICE pt_dev;
uint32 pt (uint32 dev, uint32 op, uint32 dat);
t_stat ptr_svc (UNIT *uptr);
t_stat ptp_svc (UNIT *uptr);
t_stat pt_boot (int32 unitno, DEVICE *dptr);
t_stat pt_reset (DEVICE *dptr);

/* PT data structures

   pt_dev	PT device descriptor
   pt_unit	PT unit descriptors
   pt_reg	PT register list
*/

DIB pt_dib = { d_PT, -1, v_PT, NULL, &pt, NULL };

UNIT pt_unit[] = {
	{ UDATA (&ptr_svc, UNIT_SEQ+UNIT_ATTABLE+UNIT_ROABLE, 0),
		SERIAL_IN_WAIT },
	{ UDATA (&ptp_svc, UNIT_SEQ+UNIT_ATTABLE, 0),
		SERIAL_OUT_WAIT }  };

REG pt_reg[] = {
	{ HRDATA (STA, pt_sta, 8) },
	{ HRDATA (RBUF, pt_unit[PTR].buf, 8) },
	{ DRDATA (RPOS, pt_unit[PTR].pos, 32), PV_LEFT },
	{ DRDATA (RTIME, pt_unit[PTR].wait, 24), PV_LEFT },
	{ FLDATA (RSTOP_IOE, ptr_stopioe, 0) },
	{ HRDATA (PBUF, pt_unit[PTP].buf, 8) },
	{ DRDATA (PPOS, pt_unit[PTP].pos, 32), PV_LEFT },
	{ DRDATA (PTIME, pt_unit[PTP].wait, 24), PV_LEFT },
	{ FLDATA (PSTOP_IOE, ptp_stopioe, 0) },
	{ FLDATA (IREQ, int_req[l_PT], i_PT) },
	{ FLDATA (IENB, int_enb[l_PT], i_PT) },
	{ FLDATA (IARM, pt_arm, 0) },
	{ FLDATA (RD, pt_rd, 0) },
	{ FLDATA (RUN, pt_run, 0) },
	{ FLDATA (SLEW, pt_slew, 0) },
	{ FLDATA (CHP, pt_chp, 0) },
	{ HRDATA (DEVNO, pt_dib.dno, 8), REG_HRO },
	{ NULL }  };

MTAB pt_mod[] = {
	{ MTAB_XTD|MTAB_VDV, 0, "devno", "DEVNO",
		&set_dev, &show_dev, NULL },
	{ 0 }  };

DEVICE pt_dev = {
	"PT", pt_unit, pt_reg, pt_mod,
	2, 10, 31, 1, 16, 8,
	NULL, NULL, &pt_reset,
	&pt_boot, NULL, NULL,
	&pt_dib, DEV_DISABLE };

/* Paper tape: IO routine */

uint32 pt (uint32 dev, uint32 op, uint32 dat)
{
uint32 t, old_rd, old_run;

switch (op) {						/* case IO op */
case IO_ADR:						/* select */
	return BY;					/* byte only */
case IO_OC:						/* command */
	old_rd = pt_rd;					/* save curr rw */
	old_run = pt_run;				/* save curr run */
	pt_arm = int_chg (v_PT, dat, pt_arm);		/* upd int ctrl */
	pt_rd = io_2b (dat, CMD_V_RD, pt_rd);		/* upd read/wr */
	if (old_rd != pt_rd) {				/* rw change? */
	    pt_sta = pt_sta & ~STA_OVR;			/* clr overrun */
	    if (sim_is_active (&pt_unit[pt_rd? PTR: PTP])) {
		pt_sta = pt_sta | STA_BSY;		/* busy = 1 */
		CLR_INT (v_PT);  }			/* clear int */
	    else {					/* not active */
		pt_sta = pt_sta & ~STA_BSY;		/* busy = 0 */
		if (pt_arm) SET_INT (v_PT);  }  }	/* no, set int */
	if (pt_rd) {					/* reader? */
	    pt_run = io_2b (dat, CMD_V_RUN, pt_run);	/* upd run/stop */
	    pt_slew = io_2b (dat, CMD_V_SLEW, pt_slew);	/* upd slew/inc */
	    if (pt_run) {				/* run set? */
		if (old_run == 0) {			/* run 0 -> 1? */
		    sim_activate (&pt_unit[PTR], pt_unit[PTR].wait);
		    pt_sta = pt_sta & ~STA_DU;  }  }	/* clear eof */
	    else sim_cancel (&pt_unit[PTR]);  }		/* clr, stop rdr */
	else pt_sta = pt_sta & ~STA_DU;			/* punch, clr eof */
	break;
case IO_RD:						/* read */
	if (pt_run && !pt_slew) {			/* incremental? */
	    sim_activate (&pt_unit[PTR], pt_unit[PTR].wait);
	    pt_sta = pt_sta & ~STA_DU;  }		/* clr eof */
	pt_chp = 0;					/* clr char pend */
	if (pt_rd) pt_sta = pt_sta | STA_BSY;		/* set busy */
	return (pt_unit[PTR].buf & 0xFF);		/* return char */
case IO_WD:						/* write */
	pt_unit[PTP].buf = dat & DMASK8;		/* save char */
	if (!pt_rd) pt_sta = pt_sta | STA_BSY;		/* set busy */
	sim_activate (&pt_unit[PTP], pt_unit[PTP].wait);
	break;
case IO_SS:						/* status */
	t = pt_sta & STA_MASK;				/* get status */
	if (pt_rd && !pt_run && !sim_is_active (&pt_unit[PTR]))
	    t = t | STA_NMTN;				/* stopped? */
	if ((pt_unit[pt_rd? PTR: PTP].flags & UNIT_ATT) == 0)
	    t = t | STA_DU;				/* offline? */
	if (t & SET_EX) t = t | STA_EX;			/* test for EX */
	return t;  }
return 0;
}

/* Unit service */

t_stat ptr_svc (UNIT *uptr)
{
uint32 temp;

if ((uptr->flags & UNIT_ATT) == 0)			/* attached? */
	return IORETURN (ptr_stopioe, SCPE_UNATT);
if (pt_rd) {						/* read mode? */
	pt_sta = pt_sta & ~STA_BSY;			/* clear busy */
	if (pt_arm) SET_INT (v_PT);			/* if armed, intr */
	if (pt_chp) pt_sta = pt_sta | STA_OVR;  }	/* overrun? */
pt_chp = 1;						/* char pending */
if (pt_slew) sim_activate (uptr, uptr->wait);		/* slew? continue */
if ((temp = getc (uptr->fileref)) == EOF) {		/* error? */
	if (feof (uptr->fileref)) {			/* eof? */
	    pt_sta = pt_sta | STA_DU;			/* set DU */
	    if (ptr_stopioe) printf ("PTR end of file\n");
	    else return SCPE_OK;  }
	else perror ("PTR I/O error");
	clearerr (uptr->fileref);
	return SCPE_IOERR;  }
uptr->buf = temp & DMASK8;				/* store char */
uptr->pos = uptr->pos + 1;				/* incr pos */
return SCPE_OK;
}

t_stat ptp_svc (UNIT *uptr)
{
if ((uptr->flags & UNIT_ATT) == 0)			/* attached? */
	return IORETURN (ptp_stopioe, SCPE_UNATT);
if (!pt_rd) {						/* write mode? */
	pt_sta = pt_sta & ~STA_BSY;			/* clear busy */
	if (pt_arm) SET_INT (v_PT);  }			/* if armed, intr */
if (putc (uptr->buf, uptr -> fileref) == EOF) {		/* write char */
	perror ("PTP I/O error");
	clearerr (uptr -> fileref);
	return SCPE_IOERR;  }
uptr -> pos = uptr -> pos + 1;				/* incr pos */
return SCPE_OK;
}

/* Bootstrap routine */

#define BOOT_DEV	0x78
#define BOOT_IOC	0x79
#define BOOT_START	0x50
#define BOOT_LEN	(sizeof (boot_rom) / sizeof (uint8))
#define BOOT3_START	0x3E
#define BOOT3_LEN	(sizeof (boot_rom) / sizeof (uint8))

static uint8 boot_rom[] = {
	0xD5, 0x00,		/* ST:	AL CF */
	0x00, 0xCF,
	0x43, 0x00,		/*	BR 80 */
	0x00, 0x80
};

static uint8 boot3_rom[] = {
	0xC8, 0x20,		/* ST:	LHI 2,80 */
	0x00, 0x80,
	0xC8, 0x30,		/*	LHI 3,1 */
	0x00, 0x01,
	0xC8, 0x40,		/*	LHI 4,CF */
	0x00, 0xCF,
	0xD3, 0xA0,		/*	LB A,78 */
	0x00, 0x78,
	0xDE, 0xA0,		/*	OC A,79 */
	0x00, 0x79,
	0x9D, 0xAE,		/* LP:	SSR A,E */
	0x42, 0xF0,		/* 	BTC F,LP */
	0x00, 0x52,
	0x9B, 0xAE,		/*	RDR A,E */
	0x08, 0xEE,		/*	LHR E,E */
	0x43, 0x30,		/*	BZ LP */
	0x00, 0x52,
	0x43, 0x00,		/*	BR STO */
	0x00, 0x6C,
	0x9D, 0xAE,		/* LP1:	SSR A,E */
	0x42, 0xF0,		/*	BTC F,LP1 */
	0x00, 0x64,
	0x9B, 0xAE,		/*	RDR A,E */
	0xD2, 0xE2,		/* STO:	STB E,0(2) */
	0x00, 0x00,
	0xC1, 0x20,		/*	BXLE 2,LP1 */
	0x00, 0x64,
	0x43, 0x00,		/*	BR 80 */
	0x00, 0x80
};

t_stat pt_boot (int32 unitno, DEVICE *dptr)
{
extern uint32 PC, dec_flgs;
extern uint16 decrom[];

if (decrom[0xD5] & dec_flgs)				/* AL defined? */
	IOWriteBlk (BOOT3_START, BOOT3_LEN, boot3_rom);	/* no, 50 seq */
else IOWriteBlk (BOOT_START, BOOT_LEN, boot_rom);	/* yes, use AL */
IOWriteB (BOOT_DEV, pt_dib.dno);			/* set dev no */
IOWriteB (BOOT_IOC, 0x99);				/* set dev cmd */
PC = BOOT_START;
return SCPE_OK;
}

/* Reset routine */

t_stat pt_reset (DEVICE *dptr)
{
sim_cancel (&pt_unit[PTR]);				/* deactivate units */
sim_cancel (&pt_unit[PTP]);
pt_rd = 1;						/* read */
pt_chp = pt_run = pt_slew = 0;				/* stop, inc, disarm */
pt_sta = STA_BSY;					/* buf empty */
CLR_INT (v_PT);						/* clear int */
CLR_ENB (v_PT);						/* disable int */
pt_arm = 0;						/* disarm int */
return SCPE_OK;
}
